#include "linalg.h"

/****************************************************************************/
/**                                                                        **/
/**                              Utility Funtions                          **/
/**                       Adapted from the f2c distribution                **/
/**                                                                        **/
/****************************************************************************/

double d_sign P2C(double *, a, double *, b)
{
  double x;
  x = *a >= 0 ? *a : - *a;
  return *b >= 0 ? x : -x;
}

double z_abs P1C(dcomplex *, z)
{
  double real, imag;
  double temp;
  real = z->r;
  imag = z->i;

  if(real < 0)
    real = -real;
  if(imag < 0)
    imag = -imag;
  if(imag > real){
    temp = real;
    real = imag;
    imag = temp;
  }
  if((real+imag) == real)
    return(real);

  temp = imag/real;
  temp = real*sqrt(1.0 + temp*temp);  /*overflow!!*/
  return(temp);
}

VOID z_div P3C(dcomplex *, c, dcomplex *, a, dcomplex *, b)
{
  double ratio, den;
  double abr, abi;

  if( (abr = b->r) < 0.)
    abr = - abr;
  if( (abi = b->i) < 0.)
    abi = - abi;
  if( abr <= abi ) {
    if(abi == 0)
      xlfail("complex division by zero");
    ratio = b->r / b->i ;
    den = b->i * (1 + ratio*ratio);
    c->r = (a->r*ratio + a->i) / den;
    c->i = (a->i*ratio - a->r) / den;
  }
  else {
    ratio = b->i / b->r ;
    den = b->r * (1 + ratio*ratio);
    c->r = (a->r + a->i*ratio) / den;
    c->i = (a->i - a->r*ratio) / den;
  }
}

VOID d_cnjg P2C(dcomplex *, r, dcomplex *, z)
{
  r->r = z->r;
  r->i = - z->i;
}

double dcabs1 P1C(dcomplex *, z)
{
  return abs(z->r) + abs(z->i);
}

VOID z_sqrt P2C(dcomplex *, r, dcomplex *, z)
{
  double mag;

  if((mag = z_abs(z)) == 0.0)
    r->r = r->i = 0.0;
  else if(z->r > 0) {
    r->r = sqrt(0.5 * (mag + z->r));
    r->i = z->i / r->r / 2;
  }
  else {
    r->i = sqrt(0.5 * (mag - z->r));
    if(z->i < 0)
      r->i = - r->i;
    r->r = z->i / r->i / 2;
  }
}

double POW2 P1C(double, x) { return x * x; }
double ABS P1C(double, x) { return abs(x); }


/****************************************************************************/
/**                                                                        **/
/**                           Level 1 BLAS routines                        **/
/**                       translated by f2c and modified                   **/
/**                                                                        **/
/****************************************************************************/

double blas_dasum P3C(int, n, double *, dx, int, incx)
{
  /* Local variables */
  int m;
  double dtemp;

  /* takes the sum of the absolute values. */
  /* jack dongarra, linpack, 3/11/78. */
  /* modified 3/93 to return if incx .le. 0. */
  /* modified 12/3/93, array(1) declarations changed to array(*) */

  /* Function Body */
  dtemp = 0.0;
  if (n <= 0 || incx <= 0)
    return dtemp;
  if (incx != 1) {
    /* code for increment not equal to 1 */
    while (n-- > 0) {
      dtemp += ABS(*dx);
      dx += incx;
    }
  }
  else {
    /* code for increment equal to 1 */

    /* clean-up loop */
    m = n % 6;
    if (m != 0) {
      n -= m;
      while (m-- > 0) {
	dtemp += ABS(*dx);
	dx++;
      }
    }
    for (; n > 0; n -= 6, dx += 6)
      dtemp +=
	ABS(dx[0]) + ABS(dx[1]) + ABS(dx[2]) +
	  ABS(dx[3]) + ABS(dx[4]) + ABS(dx[5]);
  }
  return dtemp;
}

VOID blas_daxpy P6C(int, n,
		    double, da,
		    double *, dx,
		    int, incx,
		    double *, dy,
		    int, incy)
{
  /* Local variables */
  int m;

  /* constant times a vector plus a vector. */
  /* uses unrolled loops for increments equal to one. */
  /* jack dongarra, linpack, 3/11/78. */
  /* modified 12/3/93, array(1) declarations changed to array(*) */

  /* Function Body */
  if (n <= 0 || da == 0.0)
    return;
  if (incx != 1 || incy != 1) {
    /* code for unequal increments or equal increments */
    /* not equal to 1 */
    if (incx < 0) dx -= (n - 1) * incx;
    if (incy < 0) dy -= (n - 1) * incy;
    while (n-- > 0) {
      *dy += da * *dx;
      dx += incx;
      dy += incy;
    }
  }
  else {
    /* code for both increments equal to 1 */

    /* clean-up loop */
    m = n % 4;
    if (m != 0) {
      n -= m;
      while (m-- > 0)
	*dy++ += da * *dx++;
    }
    for (; n > 0; n -= 4, dx += 4, dy += 4) {
      dy[0] += da * dx[0];
      dy[1] += da * dx[1];
      dy[2] += da * dx[2];
      dy[3] += da * dx[3];
    }
  }
}

VOID blas_dcopy P5C(int, n, double *, dx, int, incx, double *, dy, int, incy)
{
  /* Local variables */
  int m;

  /* copies a vector, x, to a vector, y. */
  /* uses unrolled loops for increments equal to one. */
  /* jack dongarra, linpack, 3/11/78. */
  /* modified 12/3/93, array(1) declarations changed to array(*) */

  /* Function Body */
  if (n <= 0)
    return;
  if (incx != 1 || incy != 1) {
    /* code for unequal increments or equal increments not equal to 1 */
    if (incx < 0) dx -= (n - 1) * incx;
    if (incy < 0) dy -= (n - 1) * incy;
    while (n-- > 0) {
      *dy = *dx;
      dx += incx;
      dy += incy;
    }
  }
  else {
    /* code for both increments equal to 1 */

    /* clean-up loop */
    m = n % 7;
    if (m != 0) {
      n -= m;
      while (m-- > 0)
	*dy++ = *dx++;
    }
    for (; n > 0; n -= 7, dx += 7, dy += 7) {
      dy[0] = dx[0];
      dy[1] = dx[1];
      dy[2] = dx[2];
      dy[3] = dx[3];
      dy[4] = dx[4];
      dy[5] = dx[5];
      dy[6] = dx[6];
    }
  }
}

double blas_ddot P5C(int, n, double *, dx, int, incx, double *, dy, int, incy)
{
  /* Local variables */
  int m;
  double dtemp;

  /* forms the dot product of two vectors. */
  /* uses unrolled loops for increments equal to one. */
  /* jack dongarra, linpack, 3/11/78. */
  /* modified 12/3/93, array(1) declarations changed to array(*) */

  /* Function Body */
  dtemp = 0.0;
  if (n <= 0)
    return dtemp;
  if (incx != 1 || incy != 1) {
    /* code for unequal increments or equal increments not equal to 1 */
    if (incx < 0) dx -= (n - 1) * incx;
    if (incy < 0) dy -= (n - 1) * incy;
    while (n-- > 0) {
      dtemp += *dx * *dy;
      dx += incx;
      dy += incy;
    }
  }
  else {
    /* code for both increments equal to 1 */

    /* clean-up loop */
    m = n % 5;
    if (m != 0) {
      n -= m;
      while (m-- > 0)
	dtemp += *dx++ * *dy++;
    }
    for (; n > 0; n -= 5, dx += 5, dy += 5)
      dtemp +=
	dx[0] * dy[0] + dx[1] * dy[1] + dx[2] * dy[2] +
	  dx[3] * dy[3] + dx[4] * dy[4];
  }
  return dtemp;
}

double blas_dnrm2 P3C(int, n, double *, x, int, incx)
{
  /* Local variables */
  double norm, scale, absxi;
  double ssq;

  /*  DNRM2 returns the euclidean norm of a vector via the function */
  /*  name, so that */

  /*     DNRM2 := sqrt( x'*x ) */

  /*  -- This version written on 25-October-1982. */
  /*     Modified on 14-October-1993 to inline the call to DLASSQ. */
  /*     Sven Hammarling, Nag Ltd. */

  /* Function Body */
  if (n < 1 || incx < 1)
    norm = 0.0;
  else if (n == 1)
    norm = ABS(*x);
  else {
    scale = 0.0;
    ssq = 1.0;

    /* The following loop is equivalent to this call to the LAPACK */
    /* auxiliary routine: */
    /* CALL DLASSQ( N, X, INCX, SCALE, SSQ ) */
    while (n-- > 0) {
      if (*x != 0.0) {
	absxi = ABS(*x);
	if (scale < absxi) {
	  ssq = ssq * POW2(scale / absxi) + 1.0;
	  scale = absxi;
	}
	else
	  ssq += POW2(absxi / scale);
      }
      x += incx;
    }
    norm = scale * sqrt(ssq);
  }

  return norm;
}

VOID blas_drot P7C(int, n,
		   double *, dx,
		   int, incx,
		   double *, dy,
		   int, incy,
		   double, c,
		   double, s)
{
  /* Local variables */
  double dtemp;

  /* applies a plane rotation. */
  /* jack dongarra, linpack, 3/11/78. */
  /* modified 12/3/93, array(1) declarations changed to array(*) */

  /* Function Body */
  if (n <= 0)
    return;
  if (incx != 1 || incy != 1) {
    /* code for unequal increments or equal increments not equal to 1 */
    if (incx < 0) dx -= (n - 1) * incx;
    if (incy < 0) dy -= (n - 1) * incy;
    while (n-- > 0) {
      dtemp = c * *dx + s * *dy;
      *dy   = c * *dy - s * *dx;
      *dx = dtemp;
      dx += incx;
      dy += incy;
    }
  }
  else {
    /* code for both increments equal to 1 */
    for (; n > 0; n--, dx++, dy++) {
      dtemp = c * *dx + s * *dy;
      *dy   = c * *dy - s * *dx;
      *dx = dtemp;
    }
  }
}

VOID blas_drotg P4C(double *, da, double *, db, double *, c, double *, s)
{
  /* Initialized data */
  static double one = 1.0;

  /* Local variables */
  double r, scale, z, roe;

  /* construct givens plane rotation. */
  /* jack dongarra, linpack, 3/11/78. */

  roe = *db;
  if (ABS(*da) > ABS(*db))
    roe = *da;
  scale = ABS(*da) + ABS(*db);
  if (scale == 0.0) {
    *c = 1.0;
    *s = 0.0;
    r = 0.0;
    z = 0.0;
  }
  else {
    double d__1 = POW2(*da / scale);
    double d__2 = POW2(*db / scale);
    r = scale * sqrt(d__1 + d__2);
    r = d_sign(&one, &roe) * r;
    *c = *da / r;
    *s = *db / r;
    z = 1.0;
    if (ABS(*da) > ABS(*db))
      z = *s;
    if (ABS(*db) >= ABS(*da) && *c != 0.0)
      z = 1.0 / *c;
  }
  *da = r;
  *db = z;
}

VOID blas_dscal P4C(int, n, double, da, double *, dx, int, incx)
{
  /* Local variables */
  int m;

  /* scales a vector by a constant. */
  /* uses unrolled loops for increment equal to one. */
  /* jack dongarra, linpack, 3/11/78. */
  /* modified 3/93 to return if incx .le. 0. */
  /* modified 12/3/93, array(1) declarations changed to array(*) */

  /* Function Body */
  if (n <= 0 || incx <= 0)
    return;
  if (incx != 1) {
    /* code for increment not equal to 1 */
    while (n-- > 0) {
      *dx = da * *dx;
      dx += incx;
    }
  }
  else {
    /* code for increment equal to 1 */

    /* clean-up loop */
    m = n % 5;
    if (m != 0) {
      n -= m;
      while (m-- > 0) {
	*dx = da * *dx;
	dx++;
      }
    }
    for (; n > 0; n -= 5, dx += 5) {
      dx[0] = da * dx[0];
      dx[1] = da * dx[1];
      dx[2] = da * dx[2];
      dx[3] = da * dx[3];
      dx[4] = da * dx[4];
    }
  }
}

VOID blas_dswap P5C(int, n, double *, dx, int, incx, double *, dy, int, incy)
{
  /* Local variables */
  int m;
  double dtemp;

  /* interchanges two vectors. */
  /* uses unrolled loops for increments equal one. */
  /* jack dongarra, linpack, 3/11/78. */
  /* modified 12/3/93, array(1) declarations changed to array(*) */

  /* Function Body */
  if (n <= 0)
    return;
  if (incx != 1 || incy != 1) {
    /* code for unequal increments or equal increments not equal to 1 */
    if (incx < 0) dx -= (n - 1) * incx;
    if (incy < 0) dy -= (n - 1) * incy;
    while (n-- > 0) {
      dtemp = *dx;
      *dx = *dy;
      *dy = dtemp;
      dx += incx;
      dy += incy;
    }
  }
  else {
    /* code for both increments equal to 1 */

    /* clean-up loop */
    m = n % 3;
    if (m != 0) {
      n -= m;
      while (m-- > 0) {
	dtemp = *dx;
	*dx = *dy;
	*dy = dtemp;
	dx++;
	dy++;
      }
    }
    for (; n > 0; n -= 3, dx += 3, dy += 3) {
      dtemp = dx[0];
      dx[0] = dy[0];
      dy[0] = dtemp;
      dtemp = dx[1];
      dx[1] = dy[1];
      dy[1] = dtemp;
      dtemp = dx[2];
      dx[2] = dy[2];
      dy[2] = dtemp;
    }
  }
}

int blas_idamax P3C(int, n, double *, dx, int, incx)
{
  /* System locals */
  int ret_val;

  /* Local variables */
  double dmax;
  int i;

  /* finds the index of element having max. absolute value. */
  /* jack dongarra, linpack, 3/11/78. */
  /* modified 3/93 to return if incx .le. 0. */
  /* modified 12/3/93, array(1) declarations changed to array(*) */

  /* Function Body */
  if (n < 1 || incx <= 0)
    return -1;
  ret_val = 0;
  if (n == 1)
    return ret_val;
  if (incx != 1) {
    /* code for increment not equal to 1 */
    dmax = ABS(*dx);
    dx += incx;
    for (i = 1; i < n; i++, dx += incx) {
      if (ABS(*dx) > dmax) {
	ret_val = i;
	dmax = ABS(*dx);
      }
    }
  }
  else {
    /* code for increment equal to 1 */
    dmax = ABS(*dx++);
    for (i = 1; i < n; i++, dx++) {
      if (ABS(*dx) > dmax) {
	ret_val = i;
	dmax = ABS(*dx);
      }
    }
  }
  return ret_val;
}

double blas_dzasum P3C(int, n, dcomplex *, zx, int, incx)
{
  /* Local variables */
  static double stemp;


  /* takes the sum of the absolute values. */
  /* jack dongarra, 3/11/78. */
  /* modified 3/93 to return if incx .le. 0. */
  /* modified 12/3/93, array(1) declarations changed to array(*) */

  /* Function Body */
  stemp = 0.0;
  if (n <= 0 || incx <= 0)
    return stemp;
  if (incx != 1) {
    /* code for increment not equal to 1 */
    if (incx < 0) zx -= (n - 1) * incx;
    while (n-- > 0) {
      stemp += dcabs1(zx);
      zx += incx;
    }
  }
  else {
    /* code for increment equal to 1 */
    while (n-- > 0)
      stemp += dcabs1(zx++);
  }
  return stemp;
}

double blas_dznrm2 P3C(int, n, dcomplex *, x, int, incx)
{
  /* Local variables */
  double temp, norm, scale;
  double ssq;

  /*  DZNRM2 returns the euclidean norm of a vector via the function */
  /*  name, so that */

  /*     DZNRM2 := sqrt( conjg( x' )*x ) */

  /*  -- This version written on 25-October-1982. */
  /*     Modified on 14-October-1993 to inline the call to ZLASSQ. */
  /*     Sven Hammarling, Nag Ltd. */

  /* Function Body */
  if (n < 1 || incx < 1)
    norm = 0.0;
  else {
    scale = 0.;
    ssq = 1.;
    /* The following loop is equivalent to this call to the LAPACK */
    /* auxiliary routine: */
    /* CALL ZLASSQ( N, X, INCX, SCALE, SSQ ) */
    while (n-- > 0) {
      if (x->r != 0.0) {
	temp = ABS(x->r);
	if (scale < temp) {
	  ssq = ssq * POW2(scale / temp) + 1.0;
	  scale = temp;
	}
	else
	  ssq += POW2(temp / scale);
      }
      if (x->i != 0.0) {
	temp = ABS(x->i);
	if (scale < temp) {
	  ssq = ssq * POW2(scale / temp) + 1.0;
	  scale = temp;
	}
	else
	  ssq += POW2(temp / scale);
      }
      x += incx;
    }
    norm = scale * sqrt(ssq);
  }

  return norm;
}

int blas_izamax P3C(int, n, dcomplex *, zx, int, incx)
{
  /* System generated locals */
  int ret_val;

  /* Local variables */
  double smax;
  int i;
  
  /* finds the index of element having max. absolute value. */
  /* jack dongarra, 1/15/85. */
  /* modified 3/93 to return if incx .le. 0. */
  /* modified 12/3/93, array(1) declarations changed to array(*) */

  /* Function Body */
  if (n < 1 || incx <= 0)
    return -1;
  ret_val = 0;
  if (n == 1)
    return ret_val;
  if (incx != 1) {
    /* code for increment not equal to 1 */
    smax = dcabs1(zx);
    zx += incx;
    for (i = 1; i < n; i++, zx += incx) {
      if (dcabs1(zx) > smax) {
	ret_val = i;
	smax = dcabs1(zx);
      }
    }
  }
  else {
    /* code for increment equal to 1 */
    smax = dcabs1(zx);
    zx++;
    for (i = 1; i < n; i++, zx++) {
      if (dcabs1(zx) > smax) {
	ret_val = i;
	smax = dcabs1(zx);
      }
    }
  }
  return ret_val;
}

VOID blas_zaxpy P6C(int, n,
		    dcomplex *, za,
		    dcomplex *, zx,
		    int, incx,
		    dcomplex *, zy,
		    int, incy)
{
  /* System locals */
  dcomplex z__1;

  /* constant times a vector plus a vector. */
  /* jack dongarra, 3/11/78. */
  /* modified 12/3/93, array(1) declarations changed to array(*) */

  /* Function Body */
  if (n <= 0)
    return;
  if (dcabs1(za) == 0.0)
    return;
  if (incx != 1 || incy != 1) {
    /* code for unequal increments or equal increments not equal to 1 */
    if (incx < 0) zx -= (n - 1) * incx;
    if (incy < 0) zy -= (n - 1) * incy;
    while (n-- > 0) {
      z__1.r = za->r * zx->r - za->i * zx->i;
      z__1.i = za->r * zx->i + za->i * zx->r;
      zy->r = zy->r + z__1.r;
      zy->i = zy->i + z__1.i;
      zx += incx;
      zy += incy;
    }
  }
  else {
    /* code for both increments equal to 1 */
    for (; n > 0; n--, zx++, zy++) {
      z__1.r = za->r * zx->r - za->i * zx->i;
      z__1.i = za->r * zx->i + za->i * zx->r;
      zy->r = zy->r + z__1.r;
      zy->i = zy->i + z__1.i;
    }
  }
}

VOID blas_zcopy P5C(int, n,
		    dcomplex *, zx,
		    int, incx,
		    dcomplex *, zy,
		    int, incy)
{
  /* copies a vector, x, to a vector, y. */
  /* jack dongarra, linpack, 4/11/78. */
  /* modified 12/3/93, array(1) declarations changed to array(*) */

  /* Function Body */
  if (n <= 0)
    return;
  if (incx != 1 || incy != 1) {
    /* code for unequal increments or equal increments not equal to 1 */
    if (incx < 0) zx -= (n - 1) * incx;
    if (incy < 0) zy -= (n - 1) * incy;
    while (n-- > 0) {
      zy->r = zx->r;
      zy->i = zx->i;
      zx += incx;
      zy += incy;
    }
  }
  else {
    /* code for both increments equal to 1 */
    while (n-- > 0) {
      zy->r = zx->r;
      zy->i = zx->i;
      zx++;
      zy++;
    }
  }
}

VOID blas_zdotc P6C(dcomplex *, ret_val,
		    int, n,
		    dcomplex *, zx,
		    int, incx,
		    dcomplex *, zy,
		    int, incy)
{
  /* System generated locals */
  dcomplex z__2, z__3;

  /* Local variables */
  dcomplex ztemp;

  /* forms the dot product of a vector. */
  /* jack dongarra, 3/11/78. */

  /* Function Body */
  ztemp.r = 0., ztemp.i = 0.;
  ret_val->r = 0.,  ret_val->i = 0.;
  if (n <= 0)
    return;
  if (incx != 1 || incy != 1) {
    /* code for unequal increments or equal increments */
    /* not equal to 1 */
    if (incx < 0) zx -= (n - 1) * incx;
    if (incy < 0) zy -= (n - 1) * incy;
    while (n-- > 0) {
      d_cnjg(&z__3, zx);
      z__2.r = z__3.r * zy->r - z__3.i * zy->i;
      z__2.i = z__3.r * zy->i + z__3.i * zy->r;
      ztemp.r += z__2.r;
      ztemp.i += z__2.i;
      zx += incx;
      zy += incy;
    }
  }
  else {
    /* code for both increments equal to 1 */
    for (; n > 0; n--, zx++, zy++) {
      d_cnjg(&z__3, zx);
      z__2.r = z__3.r * zy->r - z__3.i * zy->i;
      z__2.i = z__3.r * zy->i + z__3.i * zy->r;
      ztemp.r += z__2.r;
      ztemp.i += z__2.i;
    }
  }
  ret_val->r = ztemp.r,  ret_val->i = ztemp.i;
  return ;
}

VOID blas_zdotu P6C(dcomplex *, ret_val,
		    int, n,
		    dcomplex *, zx,
		    int, incx,
		    dcomplex *, zy,
		    int, incy)
{
  /* System generated locals */
  dcomplex z__2;

  /* Local variables */
  static dcomplex ztemp;

  /* forms the dot product of two vectors. */
  /* jack dongarra, 3/11/78. */

  /* Function Body */
  ztemp.r = 0., ztemp.i = 0.;
  ret_val->r = 0.,  ret_val->i = 0.;
  if (n <= 0)
    return;
  if (incx != 1 || incy != 1) {
    /* code for unequal increments or equal increments */
    /* not equal to 1 */
    if (incx < 0) zx -= (n - 1) * incx;
    if (incy < 0) zy -= (n - 1) * incy;
    while (n-- > 0) {
      z__2.r = zx->r * zy->r - zx->i * zy->i;
      z__2.i = zx->r * zy->i + zx->i * zy->r;
      ztemp.r += z__2.r;
      ztemp.i += z__2.i;
      zx += incx;
      zy += incy;
    }
  }
  else {
    /* code for both increments equal to 1 */
    for (; n > 0; n--, zx++, zy++) {
      z__2.r = zx->r * zy->r - zx->i * zy->i;
      z__2.i = zx->r * zy->i + zx->i * zy->r;
      ztemp.r += z__2.r;
      ztemp.i += z__2.i;
    }
  }
  ret_val->r = ztemp.r,  ret_val->i = ztemp.i;
  return ;
}

VOID blas_zdrot P7C(int, n,
		    dcomplex *, zx,
		    int, incx,
		    dcomplex *, zy,
		    int, incy,
		    double, c,
		    double, s)
{
  /* Local variables */
  dcomplex ztemp;

  /* applies a plane rotation, where the cos and sin (c and s) are */
  /* double precision and the vectors zx and zy are double complex. */
  /* jack dongarra, linpack, 3/11/78. */

  /* Function Body */
  if (n <= 0)
    return;
  if (incx != 1 || incy != 1) {
    /* code for unequal increments or equal increments not equal to 1 */
    if (incx < 0) zx -= (n - 1) * incx;
    if (incy < 0) zy -= (n - 1) * incy;
    while (n-- > 0) {
      ztemp.r = c * zx->r + s * zy->r;
      ztemp.i = c * zx->i + s * zy->i;
      zy->r = c * zy->r - s * zx->r;
      zy->i = c * zy->i - s * zx->i;
      zx->r = ztemp.r;
      zx->i = ztemp.i;
      zx += incx;
      zy += incy;
    }
  }
  else {
    /* code for both increments equal to 1 */
    for (; n > 0; n--, zx++, zy++) {
      ztemp.r = c * zx->r + s * zy->r;
      ztemp.i = c * zx->i + s * zy->i;
      zy->r = c * zy->r - s * zx->r;
      zy->i = c * zy->i - s * zx->i;
      zx->r = ztemp.r;
      zx->i = ztemp.i;
    }
  }
}

VOID blas_zdscal P4C(int, n,
		     double, da,
		     dcomplex *, zx,
		     int, incx)
{
  /* scales a vector by a constant. */
  /* jack dongarra, 3/11/78. */
  /* modified to correct problem with negative increment, 8/21/90. */

  /* Function Body */
  if (n <= 0)
    return;
  if (incx != 1) {
    /* code for increment not equal to 1 */
    if (incx < 0) zx -= (n - 1) * incx;
    while (n-- > 0) {
      zx->r = da * zx->r;
      zx->i = da * zx->i;
      zx += incx;
    }
  }
  else {
    /* code for increment equal to 1 */
    for (; n > 0; n--, zx++) {
      zx->r = da * zx->r;
      zx->i = da * zx->i;
    }
  }
}

VOID blas_zrotg P4C(dcomplex *, ca,
		    dcomplex *, cb,
		    double *, c,
		    dcomplex *, s)
{
  /* System generated locals */
  dcomplex z__1, z__2, z__3, z__4;

  /* Local variables */
  double norm;
  dcomplex alpha;
  double scale;

  if (z_abs(ca) == 0.0) {
    *c = 0.0;
    s->r = 1.0;
    s->i = 0.0;
    ca->r = cb->r;
    ca->i = cb->i;
  }
  else {
    double d__1, d__2;
    scale = z_abs(ca) + z_abs(cb);
    z__2.r = scale;
    z__2.i = 0.0;
    z_div(&z__1, ca, &z__2);
    z__4.r = scale;
    z__4.i = 0.0;
    z_div(&z__3, cb, &z__4);
    d__1 = POW2(z_abs(&z__1));
    d__2 = POW2(z_abs(&z__3));
    norm = scale * sqrt(d__1 + d__2);
    d__1 = z_abs(ca);
    z__1.r = ca->r / d__1;
    z__1.i = ca->i / d__1;
    alpha.r = z__1.r;
    alpha.i = z__1.i;
    *c = z_abs(ca) / norm;
    d_cnjg(&z__3, cb);
    z__2.r = alpha.r * z__3.r - alpha.i * z__3.i;
    z__2.i = alpha.r * z__3.i + alpha.i * z__3.r;
    z__1.r = z__2.r / norm;
    z__1.i = z__2.i / norm;
    s->r = z__1.r;
    s->i = z__1.i;
    z__1.r = norm * alpha.r;
    z__1.i = norm * alpha.i;
    ca->r = z__1.r;
    ca->i = z__1.i;
  }
}

VOID blas_zscal P4C(int, n,
		    dcomplex *, za,
		    dcomplex *, zx,
		    int, incx)
{
  /* System generated locals */
  dcomplex z__1;

  /* scales a vector by a constant. */
  /* jack dongarra, 3/11/78. */
  /* modified to correct problem with negative increment, 8/21/90. */

  /* Function Body */
  if (n <= 0)
    return;
  if (incx != 1) {
    /* code for increment not equal to 1 */
    if (incx < 0) zx -= (n - 1) * incx;
    while (n-- > 0) {
      z__1.r = za->r * zx->r - za->i * zx->i;
      z__1.i = za->r * zx->i + za->i * zx->r;
      zx->r = z__1.r;
      zx->i = z__1.i;
      zx += incx;
    }
  }
  else {
    /* code for increment equal to 1 */
    for (; n > 0; n--, zx++) {
      z__1.r = za->r * zx->r - za->i * zx->i;
      z__1.i = za->r * zx->i + za->i * zx->r;
      zx->r = z__1.r;
      zx->i = z__1.i;
    }
  }
}

VOID blas_zswap P5C(int, n,
		    dcomplex *, zx,
		    int, incx,
		    dcomplex *, zy,
		    int, incy)
{
  /* Local variables */
  dcomplex ztemp;
  
  /* interchanges two vectors. */
  /* jack dongarra, 3/11/78. */

  /* Function Body */
  if (n <= 0)
    return;
  if (incx != 1 || incy != 1) {
    /* code for unequal increments or equal increments not equal to 1 */
    if (incx < 0) zx -= (n - 1) * incx;
    if (incy < 0) zy -= (n - 1) * incy;
    while (n-- > 0) {
      ztemp.r = zx->r;
      ztemp.i = zx->i;
      zx->r = zy->r;
      zx->i = zy->i;
      zy->r = ztemp.r;
      zy->i = ztemp.i;
      zx += incx;
      zy += incy;
    }
  }
  else {
    /* code for both increments equal to 1 */
    for (; n > 0; n--, zx++, zy++) {
      ztemp.r = zx->r;
      ztemp.i = zx->i;
      zx->r = zy->r;
      zx->i = zy->i;
      zy->r = ztemp.r;
      zy->i = ztemp.i;
    }
  }
}


/****************************************************************************/
/**                                                                        **/
/**                 XLISP-STAT interface to BLAS routines                  **/
/**                                                                        **/
/****************************************************************************/

#define gettvectype(x) ((unsigned char *) (gettvecdata(x)))[gettlength(x)]

enum {
  CD_CHARACTER = 0,
  CD_FIXTYPE,
  CD_FLOTYPE,
  CD_CXFIXTYPE,
  CD_CXFLOTYPE,
  CD_CHAR,
  CD_SHORT,
  CD_INT,
  CD_LONG,
  CD_FLOAT,
  CD_DOUBLE,
  CD_COMPLEX,
  CD_DCOMPLEX,
  CD_TRUE
};

LOCAL VOID getblasddata P3C(int, n, double **, pdx, int *, pincr)
{
  LVAL x;
  int size, type, offset, incr;

  x = xlgatvec();
  size = gettvecsize(x);
  type = gettvectype(x);
  offset = getfixnum(xlgafixnum());
  incr = getfixnum(xlgafixnum());

  if (size < offset + (n - 1) * incr + 1)
    xlerror("incompatible with access indices", x);

  switch(type) {
  case CD_FLOTYPE:
  case CD_DOUBLE:
    break;
  default:
    xlbadtype(x);
  }

  *pdx = ((double *) gettvecdata(x)) + offset;
  *pincr = incr;
}

LOCAL VOID getblaszdata P3C(int, n, dcomplex **, pzx, int *, pincr)
{
  LVAL x;
  int size, type, offset, incr;

  x = xlgatvec();
  size = gettvecsize(x);
  type = gettvectype(x);
  offset = getfixnum(xlgafixnum());
  incr = getfixnum(xlgafixnum());

  if (size < offset + (n - 1) * incr + 1)
    xlerror("incompatible with access indices", x);

  switch(type) {
  case CD_CXFLOTYPE:
  case CD_DCOMPLEX:
    break;
  default:
    xlbadtype(x);
  }

  *pzx = ((dcomplex *) gettvecdata(x)) + offset;
  *pincr = incr;
}

LOCAL VOID getdcomplex P1C(dcomplex *, z)
{
  LVAL a = xlgetarg();
  switch(ntype(a)) {
  case FIXNUM:
    z->r = (FLOTYPE) getfixnum(a);
    z->i = 0.0;
    break;
  case FLONUM:
    z->r = getflonum(a);
    z->i = 0.0;
    break;
  case COMPLEX:
    z->r = makefloat(realpart(a));
    z->i = makefloat(imagpart(a));
    break;
  default:
    xlbadtype(a);
  }
}

LVAL xblasdasum(V)
{
  int n, incx;
  double *dx;

  n = getfixnum(xlgafixnum());
  getblasddata(n, &dx, &incx);
  xllastarg();

  return cvflonum((FLOTYPE) blas_dasum(n, dx, incx));
}

LVAL xblasdaxpy(V)
{
  int n, incx, incy;
  double a, *dx, *dy;

  n = getfixnum(xlgafixnum());
  a = makefloat(xlgetarg());
  getblasddata(n, &dx, &incx);
  getblasddata(n, &dy, &incy);
  xllastarg();

  blas_daxpy(n, a, dx, incx, dy, incy);
  return NIL;
}

LVAL xblasdcopy(V)
{
  int n, incx, incy;
  double *dx, *dy;

  n = getfixnum(xlgafixnum());
  getblasddata(n, &dx, &incx);
  getblasddata(n, &dy, &incy);
  xllastarg();

  blas_dcopy(n, dx, incx, dy, incy);
  return NIL;
}

LVAL xblasddot(V)
{
  int n, incx, incy;
  double *dx, *dy;

  n = getfixnum(xlgafixnum());
  getblasddata(n, &dx, &incx);
  getblasddata(n, &dy, &incy);
  xllastarg();

  return cvflonum((FLOTYPE) blas_ddot(n, dx, incx, dy, incy));
}

LVAL xblasdnrm2(V)
{
  int n, incx;
  double *dx;

  n = getfixnum(xlgafixnum());
  getblasddata(n, &dx, &incx);
  xllastarg();

  return cvflonum((FLOTYPE) blas_dnrm2(n, dx, incx));
}

LVAL xblasdrot(V)
{
  int n, incx, incy;
  double *dx, *dy, c, s;

  n = getfixnum(xlgafixnum());
  getblasddata(n, &dx, &incx);
  getblasddata(n, &dy, &incy);
  c = makefloat(xlgetarg());
  s = makefloat(xlgetarg());
  xllastarg();

  blas_drot(n, dx, incx, dy, incy, c, s);
  return NIL;
}

LVAL xblasdrotg(V)
{
  double a, b, c, s;

  a = makefloat(xlgetarg());
  b = makefloat(xlgetarg());
  xllastarg();

  blas_drotg(&a, &b, &c, &s);

  xlnumresults = 0;
  xlresults[xlnumresults++] = cvflonum((FLOTYPE) a);
  xlresults[xlnumresults++] = cvflonum((FLOTYPE) b);
  xlresults[xlnumresults++] = cvflonum((FLOTYPE) c);
  xlresults[xlnumresults++] = cvflonum((FLOTYPE) s);
  return xlresults[0];
}

LVAL xblasdscal(V)
{
  int n, incx;
  double a, *dx;

  n = getfixnum(xlgafixnum());
  a = makefloat(xlgetarg());
  getblasddata(n, &dx, &incx);
  xllastarg();

  blas_dscal(n, a, dx, incx);
  return NIL;
}

LVAL xblasdswap(V)
{
  int n, incx, incy;
  double *dx, *dy;

  n = getfixnum(xlgafixnum());
  getblasddata(n, &dx, &incx);
  getblasddata(n, &dy, &incy);
  xllastarg();

  blas_dswap(n, dx, incx, dy, incy);
  return NIL;
}

LVAL xblasidamax(V)
{
  int n, incx;
  double *dx;

  n = getfixnum(xlgafixnum());
  getblasddata(n, &dx, &incx);
  xllastarg();

  return cvfixnum((FIXTYPE) blas_idamax(n, dx, incx));
}

LVAL xblasdzasum(V)
{
  int n, incx;
  dcomplex *zx;

  n = getfixnum(xlgafixnum());
  getblaszdata(n, &zx, &incx);
  xllastarg();

  return cvflonum((FLOTYPE) blas_dzasum(n, zx, incx));
}

LVAL xblasdznrm2(V)
{
  int n, incx;
  dcomplex *zx;

  n = getfixnum(xlgafixnum());
  getblaszdata(n, &zx, &incx);
  xllastarg();

  return cvflonum((FLOTYPE) blas_dznrm2(n, zx, incx));
}

LVAL xblasizamax(V)
{
  int n, incx;
  dcomplex *zx;

  n = getfixnum(xlgafixnum());
  getblaszdata(n, &zx, &incx);
  xllastarg();

  return cvfixnum((FIXTYPE) blas_izamax(n, zx, incx));
}

LVAL xblaszaxpy(V)
{
  int n, incx, incy;
  dcomplex za, *zx, *zy;

  n = getfixnum(xlgafixnum());
  getdcomplex(&za);
  getblaszdata(n, &zx, &incx);
  getblaszdata(n, &zy, &incy);
  xllastarg();

  blas_zaxpy(n, &za, zx, incx, zy, incy);
  return NIL;
}

LVAL xblaszcopy(V)
{
  int n, incx, incy;
  dcomplex *zx, *zy;

  n = getfixnum(xlgafixnum());
  getblaszdata(n, &zx, &incx);
  getblaszdata(n, &zy, &incy);
  xllastarg();

  blas_zcopy(n, zx, incx, zy, incy);
  return NIL;
}

LVAL xlbaszdotc(V)
{
  int n, incx, incy;
  dcomplex *zx, *zy, val;

  n = getfixnum(xlgafixnum());
  getblaszdata(n, &zx, &incx);
  getblaszdata(n, &zy, &incy);
  xllastarg();

  blas_zdotc(&val, n, zx, incx, zy, incy);
  return newdcomplex(val.r, val.i);
}

LVAL xlbaszdotu(V)
{
  int n, incx, incy;
  dcomplex *zx, *zy, val;

  n = getfixnum(xlgafixnum());
  getblaszdata(n, &zx, &incx);
  getblaszdata(n, &zy, &incy);
  xllastarg();

  blas_zdotu(&val, n, zx, incx, zy, incy);
  return newdcomplex(val.r, val.i);
}

LVAL xlbaszdrot(V)
{
  int n, incx, incy;
  dcomplex *zx, *zy;
  double c, s;

  n = getfixnum(xlgafixnum());
  getblaszdata(n, &zx, &incx);
  getblaszdata(n, &zy, &incy);
  c = makefloat(xlgetarg());
  s = makefloat(xlgetarg());
  xllastarg();

  blas_zdrot(n, zx, incx, zy, incy, c, s);
  return NIL;
}

LVAL xlbaszdscal(V)
{
  int n, incx;
  dcomplex *zx;
  double da;

  n = getfixnum(xlgafixnum());
  da = makefloat(xlgetarg());
  getblaszdata(n, &zx, &incx);
  xllastarg();

  blas_zdscal(n, da, zx, incx);
  return NIL;
}

LVAL xlbaszrotg(V)
{
  dcomplex a, b, s;
  double c;

  getdcomplex(&a);
  getdcomplex(&b);
  xllastarg();

  blas_zrotg(&a, &b, &c, &s);

  xlnumresults = 0;
  xlresults[xlnumresults++] = newdcomplex(a.r, a.i);
  xlresults[xlnumresults++] = newdcomplex(b.r, b.i);
  xlresults[xlnumresults++] = cvflonum((FLOTYPE) c);
  xlresults[xlnumresults++] = newdcomplex(s.r, s.i);
  return xlresults[0];
}

LVAL xlbaszscal(V)
{
  int n, incx;
  dcomplex *zx, da;

  n = getfixnum(xlgafixnum());
  getdcomplex(&da);
  getblaszdata(n, &zx, &incx);
  xllastarg();

  blas_zscal(n, &da, zx, incx);
  return NIL;
}

LVAL xlbaszswap(V)
{
  int n, incx, incy;
  dcomplex *zx, *zy;

  n = getfixnum(xlgafixnum());
  getblaszdata(n, &zx, &incx);
  getblaszdata(n, &zy, &incy);
  xllastarg();

  blas_zswap(n, zx, incx, zy, incy);
  return NIL;
}
